import os
import numpy as np
from scipy.stats import gaussian_kde
import warnings

import pandas.plotting

# from pandas.plotting._matplotlib.tools import (_subplots, _flatten)


# from 

# try:
#     from pandas.plotting._tools import (_subplots, _flatten)
# except:
#     #TODO this is a quick fix for #38
#     from pandas.plotting._matplotlib.tools import (_subplots, _flatten)

from pandas import (DataFrame, Series)
from pandas.core.dtypes.common import is_number
from pandas.core.groupby import DataFrameGroupBy
from matplotlib import pyplot as plt
from warnings import warn

_DEBUG = False

def _x_range(data, extra=0.2):
    """ Compute the x_range, i.e., the values for which the
        density will be computed. It should be slightly larger than
        the max and min so that the plot actually reaches 0, and
        also has a bit of a tail on both sides.
    """
    try:
        sample_range = np.nanmax(data) - np.nanmin(data)
    except ValueError:
        return []
    if sample_range < 1e-6:
        return [np.nanmin(data), np.nanmax(data)]
    return np.linspace(np.nanmin(data) - extra*sample_range,
                       np.nanmax(data) + extra*sample_range, 1000)

def _setup_axis(ax, x_range, col_name=None, grid=False, ylabelsize=None, yrot=None):
    """ Setup the axis for the joyplot:
        - add the y label if required (as an ytick)
        - add y grid if required
        - make the background transparent
        - set the xlim according to the x_range
        - hide the xaxis and the spines
    """
    if col_name is not None:
        ax.set_yticks([0])
        ax.set_yticklabels([col_name], fontsize=ylabelsize, rotation=yrot)
        ax.yaxis.grid(grid)
    else:
        ax.yaxis.set_visible(False)
    ax.patch.set_alpha(0)
    ax.set_xlim([min(x_range), max(x_range)])
    ax.tick_params(axis='both', which='both', length=0, pad=10)
    ax.xaxis.set_visible(_DEBUG)
    ax.set_frame_on(_DEBUG)

def _is_numeric(x):
    """ Whether the array x is numeric. """
    return all(is_number(i) for i in x)

def _get_alpha(i, n, start=0.4, end=1.0):
    """ Compute alpha value at position i out of n """
    return start + (1 + i)*(end - start)/n

def _remove_na(l):
    """ Remove NA values. Should work for lists, arrays, series. """
    return Series(l).dropna().values

def _moving_average(a, n=3, zero_padded=False):
    """ Moving average of order n.
        If zero padded, returns an array of the same size as
        the input: the values before a[0] are considered to be 0.
        Otherwise, returns an array of length len(a) - n + 1 """
    ret = np.cumsum(a, dtype=float)
    ret[n:] = ret[n:] - ret[:-n]
    if zero_padded:
        return ret / n
    else:
        return ret[n - 1:] / n

def joyplot2(data, column=None, by=None, grid=False,
            xlabelsize=None, xrot=None, ylabelsize=None, yrot=None,
            ax=None, figsize=None,
            hist=False, bins=10,
            fade=False, ylim='max',
            fill=True, linecolor=None,
            overlap=1, background=None,
            labels=None, xlabels=True, ylabels=True,
            range_style='all',
            x_range=None,
            title=None,
            colormap=None,
            color=None,
            **kwds):
    """
    Draw joyplot of a DataFrame, or appropriately nested collection,
    using matplotlib and pandas.
    A joyplot is a stack of vertically aligned density plots / histograms.
    By default, if 'data' is a DataFrame,
    this function will plot a density plot for each column.
    This wrapper method tries to convert whatever structure is given
    to a nested collection of lists with additional information
    on labels, and use the private _joyplot function to actually
    draw theh plot.
    Parameters
    ----------
    data : DataFrame, Series or nested collection
    column : string or sequence
        If passed, will be used to limit data to a subset of columns
    by : object, optional
        If passed, used to form separate plot groups
    grid : boolean, default True
        Whether to show axis grid lines
    labels : boolean or list, default True.
        If list, must be the same size of the de
    xlabelsize : int, default None
        If specified changes the x-axis label size
    xrot : float, default None
        rotation of x axis labels
    ylabelsize : int, default None
        If specified changes the y-axis label size
    yrot : float, default None
        rotation of y axis labels
    ax : matplotlib axes object, default None
    figsize : tuple
        The size of the figure to create in inches by default
    hist : boolean, default False
    bins : integer, default 10
        Number of histogram bins to be used
    color : color or colors to be used in the plots. It can be:
        a string or anything interpretable as color by matplotib;
        a list of colors. See docs / examples for more details.
    kwds : other plotting keyword arguments
        To be passed to hist/kde plot function
    """

    if column is not None:
        if not isinstance(column, (list, np.ndarray)):
            column = [column]

    def _grouped_df_to_standard(grouped, column):
        converted = []
        labels = []
        for i, (key, group) in enumerate(grouped):
            if column is not None:
                group = group[column]
            labels.append(key)
            converted.append([_remove_na(group[c]) for c in group.columns if _is_numeric(group[c])])
            if i == 0:
                sublabels = [col for col in group.columns if _is_numeric(group[col])]
        return converted, labels, sublabels

    #################################################################
    # GROUPED
    # - given a grouped DataFrame, a group by key, or a dict of dicts of Series/lists/arrays
    # - select the required columns/Series/lists/arrays
    # - convert to standard format: list of lists of non-null arrays
    #   + extra parameters (labels and sublabels)
    #################################################################
    if isinstance(data, DataFrameGroupBy):
        grouped = data
        converted, _labels, sublabels = _grouped_df_to_standard(grouped, column)
        if labels is None:
            labels = _labels
    elif by is not None and isinstance(data, DataFrame):
        grouped = data.groupby(by)
        if column is None:
            # Remove the groupby key. It's not automatically removed by pandas.
            column = list(data.columns).remove(by)
        converted, _labels, sublabels = _grouped_df_to_standard(grouped, column)
        if labels is None:
            labels = _labels
        # If there is at least an element which is not a list of lists.. go on.
    elif isinstance(data, dict) and all(isinstance(g, dict) for g in data.values()):
        grouped = data
        if labels is None:
            labels = list(grouped.keys())
        converted = []
        for i, (key, group) in enumerate(grouped.items()):
            if column is not None:
                converted.append([_remove_na(g) for k,g in group.items() if _is_numeric(g) and k in column])
                if i == 0:
                    sublabels = [k for k,g in group.items() if _is_numeric(g)]
            else:
                converted.append([_remove_na(g) for k,g in group.items() if _is_numeric(g)])
                if i == 0:
                    sublabels = [k for k,g in group.items() if _is_numeric(g)]
    #################################################################
    # PLAIN:
    # - given a DataFrame or list/dict of Series/lists/arrays
    # - select the required columns/Series/lists/arrays
    # - convert to standard format: list of lists of non-null arrays + extra parameter (labels)
    #################################################################
    elif isinstance(data, DataFrame):
        if column is not None:
            data = data[column]
        converted = [[_remove_na(data[col])] for col in data.columns if _is_numeric(data[col])]
        labels = [col for col in data.columns if _is_numeric(data[col])]
        sublabels = None
    elif isinstance(data, dict):
        if column is not None:
            converted = [[_remove_na(g)] for k,g in data.items() if _is_numeric(g) and k in column]
            labels = [k for k,g in data.items() if _is_numeric(g) and k in column]
        else:
            converted = [[_remove_na(g)] for k,g in data.items() if _is_numeric(g)]
            labels = [k for k,g in data.items() if _is_numeric(g)]
        sublabels = None
    elif isinstance(data, list):
        if column is not None:
            converted = [[_remove_na(g)] for g in data if _is_numeric(g) and i in column]
        else:
            converted = [[_remove_na(g)] for g in data if _is_numeric(g)]
        if labels and len(labels) != len(converted):
            raise ValueError("The number of labels does not match the length of the list.")

        sublabels = None
    else:
        raise TypeError("Unknown type for 'data': {!r}".format(type(data)))

    if ylabels is False:
        labels = None

    if all(len(subg)==0 for g in converted for subg in g):
        raise ValueError("No numeric values found. Joyplot requires at least a numeric column/group.")

    if any(len(subg)==0 for g in converted for subg in g):
        warn("At least a column/group has no numeric values.")


    return _joyplot(converted, labels=labels, sublabels=sublabels,
                    grid=grid,
                    xlabelsize=xlabelsize, xrot=xrot, ylabelsize=ylabelsize, yrot=yrot,
                    ax=ax, figsize=figsize,
                    hist=hist, bins=bins,
                    fade=fade, ylim=ylim,
                    fill=fill, linecolor=linecolor,
                    overlap=overlap, background=background,
                    xlabels=xlabels,
                    range_style=range_style, x_range=x_range,
                    title=title,
                    colormap=colormap,
                    color=color,
                    **kwds)

###########################################

def plot_density(ax, x_range, v, kind="kde", bw_method=None,
                 bins=50,
                 fill=False, linecolor=None, clip_on=True, **kwargs):
    """ Draw a density plot given an axis, an array of values v and an array
        of x positions where to return the estimated density.
    """
    v = _remove_na(v)
    if len(v) == 0 or len(x_range) == 0:
        return

    if kind == "kde":
        try:
            gkde = gaussian_kde(v, bw_method=bw_method)
            y = gkde.evaluate(x_range)
        except ValueError:
            # Handle cases where there is no data in a group.
            y = np.zeros_like(x_range)
        except np.linalg.LinAlgError as e:
            # Handle singular matrix in kde computation.
            distinct_values = np.unique(v)
            if len(distinct_values) == 1:
                # In case of a group with a single value val,
                # that should have infinite density,
                # return a δ(val)
                val = distinct_values[0]
                warnings.warn("The data contains a group with a single distinct value ({}) "
                              "having infinite probability density. "
                              "Consider using a different visualization.".format(val))

                # Find index i of x_range
                # such that x_range[i-1] < val ≤ x_range[i]
                i = np.searchsorted(x_range, val)

                y = np.zeros_like(x_range)
                y[i] = 1
            else:
                raise e

    elif kind == "counts":
        y, bin_edges = np.histogram(v, bins=bins, range=(min(x_range), max(x_range)))
        # np.histogram returns the edges of the bins.
        # We compute here the middle of the bins.
        x_range = _moving_average(bin_edges, 2)
    elif kind == "normalized_counts":
        y, bin_edges = np.histogram(v, bins=bins, density=False,
                                    range=(min(x_range), max(x_range)))
        # np.histogram returns the edges of the bins.
        # We compute here the middle of the bins.
        y = y / len(v)
        x_range = _moving_average(bin_edges, 2)
    elif kind == "values":
        # Warning: to use values and get a meaningful visualization,
        # x_range must also be manually set in the main function.
        y = v
        x_range = list(range(len(y)))
    else:
        raise NotImplementedError

    if fill:
        ax.fill_between(x_range, 0.0, y, clip_on=clip_on, **kwargs)

        # Hack to have a border at the bottom at the fill patch
        # (of the same color of the fill patch)
        # so that the fill reaches the same bottom margin as the edge lines
        # with y value = 0.0
        kw = kwargs
        kw["label"] = None
        ax.plot(x_range, [0.0]*len(x_range), clip_on=clip_on, **kw)

    if linecolor is not None:
        kwargs["color"] = linecolor

    # Remove the legend labels if we are plotting filled curve:
    # we only want one entry per group in the legend (if shown).
    if fill:
        kwargs["label"] = None

    ax.plot(x_range, y, clip_on=clip_on, **kwargs)

###########################################

def _joyplot(data,
             grid=False,
             labels=None, sublabels=None,
             xlabels=True,
             xlabelsize=None, xrot=None,
             ylabelsize=None, yrot=None,
             ax=None, figsize=None,
             hist=False, bins=10,
             fade=False,
             xlim=None, ylim='max',
             fill=True, linecolor=None,
             overlap=1, background=None,
             range_style='all', x_range=None, tails=0.2,
             title=None,
             legend=False, loc="upper right",
             colormap=None, color=None,
             **kwargs):
    """
    Internal method.
    Draw a joyplot from an appropriately nested collection of lists
    using matplotlib and pandas.
    Parameters
    ----------
    data : DataFrame, Series or nested collection
    grid : boolean, default True
        Whether to show axis grid lines
    labels : boolean or list, default True.
        If list, must be the same size of the de
    xlabelsize : int, default None
        If specified changes the x-axis label size
    xrot : float, default None
        rotation of x axis labels
    ylabelsize : int, default None
        If specified changes the y-axis label size
    yrot : float, default None
        rotation of y axis labels
    ax : matplotlib axes object, default None
    figsize : tuple
        The size of the figure to create in inches by default
    hist : boolean, default False
    bins : integer, default 10
        Number of histogram bins to be used
    kwarg : other plotting keyword arguments
        To be passed to hist/kde plot function
    """

    if fill is True and linecolor is None:
        linecolor = "k"

    if sublabels is None:
        legend = False

    def _get_color(i, num_axes, j, num_subgroups):
        if isinstance(color, list):
            return color[j] if num_subgroups > 1 else color[i]
        elif color is not None:
            return color
        elif isinstance(colormap, list):
            return colormap[j](i/num_axes)
        elif color is None and colormap is None:
            num_cycle_colors = len(plt.rcParams['axes.prop_cycle'].by_key()['color'])
            return plt.rcParams['axes.prop_cycle'].by_key()['color'][j % num_cycle_colors]
        else:
            return colormap(i/num_axes)

    ygrid = (grid is True or grid == 'y' or grid == 'both')
    xgrid = (grid is True or grid == 'x' or grid == 'both')

    num_axes = len(data)

    if x_range is None:
        global_x_range = _x_range([v for g in data for sg in g for v in sg])
    else:
        global_x_range = _x_range(x_range, 0.0)
    global_x_min, global_x_max = min(global_x_range), max(global_x_range)

    # Each plot will have its own axis
    fig, axes = plt.subplots(num_axes, sharex=True, sharey=False,
                             figsize=figsize)
    _axes = axes.ravel()

    # The legend must be drawn in the last axis if we want it at the bottom.
    if loc in (3, 4, 8) or 'lower' in str(loc):
        legend_axis = num_axis - 1
    else:
        legend_axis = 0

    # A couple of simple checks.
    if labels is not None:
        assert len(labels) == num_axes
    if sublabels is not None:
        assert all(len(g) == len(sublabels) for g in data)
    #if isinstance(color, list):
       #assert all(len(g) <= len(color) for g in data)
    if isinstance(colormap, list):
        assert all(len(g) == len(colormap) for g in data)

    for i, group in enumerate(data):
        a = _axes[i]
        group_zorder = i
        if fade:
            kwargs['alpha'] = _get_alpha(i, num_axes)

        num_subgroups = len(group)

        if hist:
            # matplotlib hist() already handles multiple subgroups in a histogram
            a.hist(group, label=sublabels, bins=bins, color=color,
                   range=[min(global_x_range), max(global_x_range)],
                   edgecolor=linecolor, zorder=group_zorder, **kwargs)
        else:
            for j, subgroup in enumerate(group):

                # Compute the x_range of the current plot
                if range_style == 'all':
                # All plots have the same range
                    x_range = global_x_range
                elif range_style == 'own':
                # Each plot has its own range
                    x_range = _x_range(subgroup, tails)
                elif range_style == 'group':
                # Each plot has a range that covers the whole group
                    x_range = _x_range(group, tails)
                elif isinstance(range_style, (list, np.ndarray)):
                # All plots have exactly the range passed as argument
                    x_range = _x_range(range_style, 0.0)
                else:
                    raise NotImplementedError("Unrecognized range style.")

                if sublabels is None:
                    sublabel = None
                else:
                    sublabel = sublabels[j]

                element_zorder = group_zorder + j/(num_subgroups+1)
                element_color = _get_color(i, num_axes, j, num_subgroups)

                plot_density(a, x_range, subgroup,
                             fill=fill, linecolor=linecolor, label=sublabel,
                             zorder=element_zorder, color=element_color,
                             bins=bins, **kwargs)


        # Setup the current axis: transparency, labels, spines.
        col_name = None if labels is None else labels[i]
        _setup_axis(a, global_x_range, col_name=col_name, grid=ygrid,
                ylabelsize=ylabelsize, yrot=yrot)

        # When needed, draw the legend
        if legend and i == legend_axis:
            a.legend(loc=loc)
            # Bypass alpha values, in case
            for p in a.get_legend().get_patches():
                p.set_facecolor(p.get_facecolor())
                p.set_alpha(1.0)
            for l in a.get_legend().get_lines():
                l.set_alpha(1.0)


    # Final adjustments

    # Set the y limit for the density plots.
    # Since the y range in the subplots can vary significantly,
    # different options are available.
    if ylim == 'max':
        # Set all yaxis limit to the same value (max range among all)
        max_ylim = max(a.get_ylim()[1] for a in _axes)
        min_ylim = min(a.get_ylim()[0] for a in _axes)
        for a in _axes:
            a.set_ylim([min_ylim - 0.1*(max_ylim-min_ylim), max_ylim])

    elif ylim == 'own':
        # Do nothing, each axis keeps its own ylim
        pass

    else:
        # Set all yaxis lim to the argument value ylim
        try:
            for a in _axes:
                a.set_ylim(ylim)
        except:
            print("Warning: the value of ylim must be either 'max', 'own', or a tuple of length 2. The value you provided has no effect.")

    # Compute a final axis, used to apply global settings
    last_axis = fig.add_subplot(1, 1, 1)

    # Background color
    if background is not None:
        last_axis.patch.set_facecolor(background)

    for side in ['top', 'bottom', 'left', 'right']:
        last_axis.spines[side].set_visible(_DEBUG)

    # This looks hacky, but all the axes share the x-axis,
    # so they have the same lims and ticks
    last_axis.set_xlim(_axes[0].get_xlim())
    if xlabels is True:
        last_axis.set_xticks(np.array(_axes[0].get_xticks()[1:-1]))
        for t in last_axis.get_xticklabels():
            t.set_visible(True)
            t.set_fontsize(xlabelsize)
            t.set_rotation(xrot)

        # If grid is enabled, do not allow xticks (they are ugly)
        if xgrid:
            last_axis.tick_params(axis='both', which='both',length=0)
    else:
        last_axis.xaxis.set_visible(False)

    last_axis.yaxis.set_visible(False)
    last_axis.grid(xgrid)


    # Last axis on the back
    last_axis.zorder = min(a.zorder for a in _axes) - 1
    _axes = list(_axes) + [last_axis]

    if title is not None:
        plt.title(title)


    # The magic overlap happens here.
    h_pad = 5 + (- 5*(1 + overlap))
    fig.tight_layout(h_pad=h_pad)


    return fig, _axes



# -*- coding: utf-8 -*-
"""
Quantiles Local Projections Wrapper
rlafarguette@imf.org
Time-stamp: "2021-09-24 15:09:05 RLafarguette"
"""

###############################################################################
#%% Import
###############################################################################
# Base
import os, sys, importlib                               # System packages
import pandas as pd                                     # Dataframes
import numpy as np                                      # Numeric tools
import statsmodels as sm                                # Statistical models
import statsmodels.formula.api as smf                   # Formulas

# Functional imports
from scipy import interpolate                           # Linear interpolation
from scipy.stats import t  # Tdistribution
from scipy.stats import norm  # Gaussian distribution
from collections import namedtuple                      # High perf container

# Plotting packages
import matplotlib.pyplot as plt
import seaborn as sns

# Warnings management
# With many quantile regressions, the convergence warnings are overwhelming
from  warnings import simplefilter                       # Filter warnings

from statsmodels.tools.sm_exceptions import (ConvergenceWarning,
                                             IterationLimitWarning)
simplefilter("ignore", category=ConvergenceWarning)
simplefilter("ignore", category=IterationLimitWarning)

# Specific warnings in quantile regressions 
np.seterr(divide='ignore', invalid='ignore')

###############################################################################
#%% Ancillary functions
###############################################################################
def zscore(series):
    """ Zscore a pandas series """
    return((series - series.mean())/series.std(ddof=0))

# Quantile Interpolation following Schmidt and Zhu (2016)
def quantile_interpolation(alpha, cond_quant_dict):
    """ 
    Quantile interpolation function, following Schmidt and Zhu (2016) p12
    - Alpha is the quantile that needs to be interpolated
    - cond_quant_dict is the dictionary of quantiles to interpolate on 
    Return:
    - The interpolated quantile
    """

    # List of quantiles
    qlist = sorted(list(cond_quant_dict.keys()))
    min_q = min(qlist)
    max_q = max(qlist)

    # Fix the base quantile function (usually a N(0,1))
    base = norm.ppf

    # Considering multiple cases
    if alpha in qlist: ## No need to interpolate, just on the spot !!
        interp = cond_quant_dict[alpha]

    elif alpha < min_q: ## The left edge
        ## Compute the slope (page 13) 
        b1_up = (cond_quant_dict[max_q] - cond_quant_dict[min_q])
        b1_low = base(max_q) - base(min_q)
        b1 = b1_up/b1_low

        ## Compute the intercept (page 12)
        a1 = cond_quant_dict[min_q] - b1*base(min_q)

        ## Compute the interpolated value
        interp = a1 + b1*base(alpha)
        
    elif alpha > max_q: # The right edge (same formula)
        # Compute the slope (page 13) 
        b1_up = (cond_quant_dict[max_q] - cond_quant_dict[min_q])
        b1_low = base(max_q) - base(min_q)
        b1 = b1_up/b1_low

        # Compute the intercept (page 12)
        a1 = cond_quant_dict[min_q] - b1*base(min_q)

        # Compute the interpolated value
        interp = a1 + b1*base(alpha)

    else: # In the belly
        # Need to identify the closest quantiles
        local_min_list = [x for x in qlist if x < alpha]
        local_min = max(local_min_list) # The one immediately below

        local_max_list = [x for x in qlist if x > alpha]
        local_max = min(local_max_list) # The one immediately above

        # Compute the slope
        b_up = (cond_quant_dict[local_max] - cond_quant_dict[local_min])
        b_low = base(local_max) - base(local_min)
        b = b_up/b_low

        # Compute the intercept
        a = cond_quant_dict[local_max] - b*base(local_max)
        
        # Compute the interpolated value
        interp = a + b*base(alpha)

    # Return the interpolated quantile    
    return(interp)


# Uncrossing using Cherzonukov et al (Econometrica 2010)
def quantiles_uncrossing(cond_quant_dict, method='linear', len_bs=1000,
                         seed=None):
    """ 
    Uncross a set of conditional_quantiles either via Cherzonukov et al (2010)
    (bootstrapped rearrangement) or Schmidt and Zhu (functional interpolation)
    
    Input:
    - A dictionary of quantile: conditional quantiles
    - Interpolation method: either linear or probabilistic. 
    The probabilistic quantile interpolation follows Zhu and Schmidt 2016
    - len_bs: length of the bootstrapped rearrangement
    Output:
    - A dictionary of quantile: uncrossed conditional quantiles
    """
    __description = "Quantiles uncrossing, following Cherzonukov et al. (2010)"
    __author = "Romain Lafarguette, IMF, rlafarguette@imf.org"
    
    ## List of quantiles
    ql = sorted(list(cond_quant_dict.keys()))
    
    cond_quant = [cond_quant_dict[q] for q in ql] # Because dict is not ordered

    # Treatment of the seed for the random numbers generator
    if seed: # If seed fixed, then use it
        np.random.seed(seed)
    else: # Else, use default
        pass
    
    # Uncrossing
    if sorted(cond_quant) == cond_quant: # Check if the quantiles are crossed
        #print('Conditional quantiles already sorted !')
        cond_quant_uncrossed_dict = cond_quant_dict
        
    else: # Uncross them using either of the two methods
        if method=='linear': # Chernozukov et al (Econometrica, 2010)        
            inter_lin = interpolate.interp1d(ql, cond_quant,
                                             fill_value='extrapolate')

            # Bootstrap the quantile function
            bootstrap_qf = inter_lin(np.random.uniform(0, 1, len_bs))

            # Now compute the percentiles of the bootstrapped quantiles 
            cond_quant_uncrossed = [np.percentile(bootstrap_qf, 100*q)
                                    for q in ql]

            # They are the uncrossed quantiles ! (super simple)
            cond_quant_uncrossed_dict = dict(zip(ql, cond_quant_uncrossed))

        elif method=='probabilistic': # Use Schmidt and Zhu (2016) approach
            bootstrap_qf = [quantile_interpolation(u, cond_quant_dict)
                            for u in np.random.uniform(0, 1, len_bs)]

            # Now compute the percentiles of the bootstrapped quantiles 
            cond_quant_uncrossed = [np.percentile(bootstrap_qf, 100*q)
                                    for q in ql]

            # They are the uncrossed quantiles !
            cond_quant_uncrossed_dict = dict(zip(ql, cond_quant_uncrossed))

        else:
            raise ValueError('Interpolation method misspecified')
            
    # Return the uncrossed quantiles    
    return(cond_quant_uncrossed_dict)

# Inverse transform sampling function
def inv_transform(cond_quant_dict, len_sample=1000, method='linear',
                  len_bs=1000, seed=None):
    """ 
    Sample a list of conditional quantiles via inverse transform sampling

    Parameters
    ----------
      cond_quant_dict : dictionary     
        Dictionary of conditional quantiles as {quantile: conditional quantile}

      quantile_list: list
        List of quantiles (should match cond_quant)

      len_sample: integer, default=1000
        Length of the sample to be returned

      len_bs=integer, default=1000
        Length of the bootstrap for quantiles uncrossing

      seed:integer, default=None
        Seed of the random numbers generator, for replicability purposes


    Returns
    -------
      A sample of length len_sample, with the quantile matching cond_quant

    """

    __description = "Inverse Transform Sampling from Conditional Quantiles"
    __author = "Romain Lafarguette, IMF, rlafarguette@imf.org"


    # Treatment of the seed for the random numbers generator
    if seed: # If seed fixed, then use it
        np.random.seed(seed)
    else: # Else, use default
        pass
        
    # Uncross the conditional quantiles if necessary
    # Rearrangement either Cherzonukov et al. (2010) or Schmidt and Zhu (2016)
    u_cond_quant_dict = quantiles_uncrossing(cond_quant_dict,
                                             method=method,
                                             len_bs=1000, seed=None)
    
    # Extract lists from dictionary
    quantile_list = sorted(u_cond_quant_dict.keys())
    num_quantile_list = [np.float(q) for q in quantile_list]
    u_cond_quant = [u_cond_quant_dict[k] for k in quantile_list] 

    # Set up an interpolation of the empirical cdf
    inv_ecdf = interpolate.interp1d(num_quantile_list, u_cond_quant,
                                    fill_value='extrapolate')

    # Draw a random sample
    U = np.random.rand(len_sample)
    
    # Use the inv_ecdf to invert a random uniform sample
    sample = inv_ecdf(U)
    
    return(sample) 


###############################################################################
#%% Parent class for the quantile projections
###############################################################################
class QuantileProj(object):
    """ 
    Specify a conditional quantile regression model

    Inputs
    ------
    depvar: string, 
       dependent variable 

    indvar_l: list
       list of independent variables. Intercept in included by default

    data: pd.DataFrame
       data to train the model on    

    """
    __description = "Quantile regressions wrapper"
    __author = "Romain Lafarguette, IMF, https://github.com/romainlafarguette"

    # Initializer
    def __init__(self, depvar, indvar_l, data,
                 horizon_l=[0], ):

        # Unit tests (defined at the bottom of the class)
        self.__quantilemod_unittest(depvar, indvar_l, data, horizon_l)
    
        # Attributes
        self.depvar = depvar # Dependent variable
        self.indvar_l = indvar_l
        self.horizon_l = sorted(horizon_l)

        # Data cleaning for the regression (no missing in dep and regressors)
        self.data = data[[self.depvar] + self.indvar_l].dropna().copy()
        
        # Print a warning in case of missing observations
        mn = data.shape[0] - self.data.shape[0]
        if mn > 0 : print(f'{mn:.0f} missing obs on depvar and indvar')
        
        # Create the forward variables based on the list of horizons
        self.depvar_l = list()
        for h in horizon_l:
            if h == 0:
                self.depvar_l.append(self.depvar)
            if h > 0:
                fname = f'{depvar}_fwd_{h}'
                self.depvar_l.append(fname)
                self.data[fname] = self.data[self.depvar].shift(-h)
                
        # Formula regressions for each dependent variable
        self.regform_d = {dv: self.__reg_formula(dv) for dv in self.depvar_l}
        
        # Run in parallel a zscore version of the data
        self.zdata = self.data.apply(zscore, axis=1).copy()

        
    # Class-methods (methods which returns a class defined below)    
    def fit(self, quantile_l=[0.05, 0.1, 0.25, 0.5, 0.75, 0.9, 0.95],
            alpha=0.1):
        """ Fit the quantile regressions for each quantile, horizon """
        return(QuantileFit(self, quantile_l, alpha))
                    
    # Methods
    def __reg_formula(self, ldepvar):
        """ Generate the specification for the quantile regressions """
        # NB: I prefer formulas as the sm output is clearer
        regressors_l = self.indvar_l[0]
        for v in self.indvar_l[1:]: regressors_l += f' + {v}'
        reg_f = f'{ldepvar} ~ {regressors_l}'
        return(reg_f)
    
    def __quantilemod_unittest(self, depvar, indvar_l, data, horizon_l):
        """ Unit testing on the inputs """
        # Test for types
        assert isinstance(depvar, str), 'depvar should be string'
        assert isinstance(indvar_l, list), 'indvars should be in list'
        assert isinstance(data, pd.DataFrame), 'data should be pandas frame'
        assert isinstance(horizon_l, list), 'horizons should be in list'

        # Types and boundaries
        for var in indvar_l:
            assert isinstance(var, str), 'each indvar should be string'
                    
        for horizon in horizon_l:
            assert isinstance(horizon, int), 'horizons should be integer'
            
        # Test for consistency
        mv_l = [x for x in [depvar] + indvar_l if x not in data.columns]
        assert len(mv_l)==0, f'{mv_l} are not in data columns'


###############################################################################
#%% Class for the quantile fit
###############################################################################
class QuantileFit(object): # Fit class for the QuantileProj class

    """ 
    Fit a the quantile regressions

    Inputs
    ------
    quantile_l: list, default [0.05, 0.1, 0.25, 0.5, 0.75, 0.9, 0.95]
       List of quantiles to fit the regressions upon
    
    alpha: float, default 0.05
       Level of confidence for the asymptotic test
              
    """

    # Import from QuantileProj class
    def __init__(self, QuantileProj, quantile_l, alpha): 
        self.__dict__.update(QuantileProj.__dict__) # Pass all attributes 
        self.__quantilefit_unittest(quantile_l, alpha) # Unit tests input

        # Attributes
        # Avoid floating point issue
        self.quantile_l = sorted([round(x,3) for x in quantile_l])
        self.alpha = alpha

        self.qfit_l = self.__qfit_l() # Return the fit of the qreg
        self.coeffs = self.__coeffs() # Very important: all the coefficients


        # Class-attributes (attributes from a class defined below)
        """ Plot the output of the fit """
        self.plot = QuantileFitPlot(self)
        
    def __qfit_l(self): 
        """ Fit a quantile regression at every quantile and horizon """

        # Prepare a container for each individual fit (convenient later)
        QFit =  namedtuple('Qfit', ['depvar', 'horizon', 'tau', 'qfit'])
        
        qfit_l = list() # Container
        
        for h, depvar in zip(self.horizon_l, self.depvar_l): 
            reg_f = self.regform_d[depvar] # Formula
                
            for tau in self.quantile_l: # For every tau

                # Estimate the quantile regression
                p = {'q':tau, 'maxiter':1000, 'p_tol':1e-05}
                qfit = smf.quantreg(formula=reg_f, data=self.data).fit(**p)

                # Package it into a container
                nt = {'depvar':depvar, 'horizon':h, 'tau':tau, 'qfit':qfit}
                qfit_l.append(QFit(**nt))
                
        print(f'{len(qfit_l)} quantile regressions estimated '
              f'for {len(self.horizon_l)} horizons '
              f'and {len(self.quantile_l)} quantiles')
        
        return(qfit_l)


    # Class-methods (methods which returns a class defined below)    
    def proj(self, cond_vector):
        """ Project quantiles based on a conditioning vector """
        return(QuantileProjection(self, cond_vector))

    
    def __coeffs(self):
        """ Create the frame of coefficients from all the quantile fit """

        depvar_frames_l = list() # Container        
        for qf in self.qfit_l:
            qfit = qf.qfit
            stats = [qfit.params, qfit.tvalues, qfit.pvalues,
                     qfit.conf_int(alpha=self.alpha)]
            
            stats_names = ['coeff', 'tval', 'pval', 'lower_ci', 'upper_ci']

            # Package it into a small dataframe
            dp = pd.concat(stats, axis=1); dp.columns = stats_names

            # Add information
            dp['pseudo_r2'] = qfit.prsquared
            dp.insert(0, 'tau', qf.tau)
            dp.insert(1, 'horizon', qf.horizon)

            # Store it
            depvar_frames_l.append(dp)

        # Concatenate all the frames to have a summary coefficients frame
        coeffs = pd.concat(depvar_frames_l)
        return(coeffs)


    # Unit tests
    def __quantilefit_unittest(self, quantile_l, alpha):
        """ Unit testing on the inputs """
        # Test for types
        assert isinstance(quantile_l, list), 'quantiles should be in list'

        # Test boundaries
        assert (0 < alpha < 1), 'level of confidence should be in (0,1)'
        for quantile in quantile_l:
            assert (0 < quantile < 1), 'quantile should be in (0,1)'

    

###############################################################################
#%% Projection class for the quantile fit class
###############################################################################
class QuantileProjection(object): # Projection class for the fit class

    """ 
    Project for a given conditioning vector

    Inputs
    ------
    cond_vector: Conditioning vector
                  
    """


    # Import from QuantileProj class
    def __init__(self, QuantileFit, cond_frame):
        self.__dict__.update(QuantileFit.__dict__) # Pass all attributes      
        self.__quantileproj_unittest(cond_frame) # Unit tests input
        
        # Attributes
        self.cond_frame = cond_frame.reset_index().copy() # In case of mult dim
        self.cond_quant = self.__proj_cond_quant()
        self.sample = self.sample # To have it as attribute for the plot class
        
        # Class-attributes (attributes from a class defined below)
        """ Plot the output of the projection """
        self.plot = QuantileProjectionPlot(self)
        
    # Methods    
    def sample(self, len_sample=1000, method='linear', len_bs=1000, seed=None):
        """ Sample from the conditional quantiles """
        
        dcq = self.cond_quant.copy() # Conditional quantile frame
        sample_frames_l = list() # Container
        for horizon in self.horizon_l:
            # Generate dictionary of conditional quantiles per horizon, sample
            for condition in sorted(set(dcq['conditioning'])):
                cond = ((dcq['horizon']==horizon) &
                        (dcq['conditioning']==condition))
                dcqh = dcq.loc[cond, :].copy()
                cq_d = {k:v for k,v in zip(dcqh['tau'],
                                           dcqh['conditional_quantile_mean'])}

                # Inverse transform sampling
                sample = inv_transform(cq_d, len_sample=len_sample,
                                       method=method, len_bs=len_bs, seed=seed)

                # Package it nicely
                ds = pd.DataFrame(sample, columns=[self.depvar])
                ds.insert(0, 'conditioning', condition) # Keep track
                ds.insert(1, 'horizon', horizon) # Add information
                sample_frames_l.append(ds)

        # Concatenate the frames    
        dsample = pd.concat(sample_frames_l)
        dsample = dsample.sort_values(by=['conditioning', 'horizon']).copy()
        return(dsample)

        
    def __proj_cond_quant(self):
        """ Project the conditional quantiles """
        
        dc_l = list() # Container
        for qf in self.qfit_l:
            qfit = qf.qfit
            dc = qfit.get_prediction(exog=self.cond_frame).summary_frame()
            dc.columns = ['conditional_quantile_' + x for x in dc.columns]
            dc = dc.set_index(self.cond_frame.index)

            # Add extra information
            dc.insert(0, 'conditioning', dc.index) # In case of mult dimension
            dc.insert(1, 'horizon', qf.horizon)
            dc.insert(2, 'tau', qf.tau)
            
            dc_l.append(dc) # Append to the container

        dcq = pd.concat(dc_l).reset_index(drop=True).copy()
        dcq = dcq.sort_values(by=['conditioning', 'horizon', 'tau']).copy()
        dcq = dcq.set_index(['conditioning', 'horizon', 'tau'],
                            drop=False).copy()
        return(dcq)
            
           
    # Unit tests
    def __quantileproj_unittest(self, cond_frame):
        """ Unit testing for the projection class """

        # Type testing
        c = isinstance(cond_frame, pd.DataFrame)
        assert c, 'cond_frame should be a pd.DataFrame with var in columns'

        # Test if the conditioning vector contains the independent variables
        mv_l = [x for x in self.indvar_l if x not in cond_frame.columns]
        assert len(mv_l)==0, f'{mv_l} not in conditioning frame columns'


###############################################################################
#%% Plot class for the QuantileFit class
###############################################################################
class QuantileFitPlot(object): # Plot class for the fit class

    """ 
    Plot the output of the different projections
    """

    # Import from QuantileProj class
    def __init__(self, QuantileFit):
        self.__dict__.update(QuantileFit.__dict__) # Pass all attributes


    # Methods    
    def coeffs_grid(self, horizon, title=None, num_cols=3, 
                     label_d={}, **kwds):
        """ 
        Plot the coefficients with confidence interval and R2 

        Parameters
        -----------        
        horizon: int
          Coefficients for the quantiles at a given horizon

        title: str, default 'Quantile Coefficients and Pseudo R2' 
          Sup title of the plot

        num_cols: int, default 3
          Number of columns, number of rows adjusts automatically

        label_d: dict, default empty
          Label dictionary to replace the subplots caption selectively

        """

        assert isinstance(horizon, int), 'Horizon should be integer'
        assert horizon in self.horizon_l, 'Horizon not in horizon list'
        
        # List of regressors
        var_l = ['Intercept'] + self.indvar_l
        total_plots = len(var_l) + 1 # add R2 square 

        # Compute the number of rows required
        num_rows = total_plots // num_cols

        if total_plots % num_cols >0:
            num_rows += 1 # Add one row if residuals charts
                
        # Line plot
        dc = self.coeffs.loc[self.coeffs['horizon']==horizon, :].copy()

        # Create the main figure
        fig, axs = plt.subplots(nrows=num_rows, ncols=num_cols, sharex=True)

        axs = axs.ravel() # Very helpful !!

        # In case, customize the labels for the plots
        label_l = [None] * len(var_l)
        
        # Replace the values in list with labels_d
        for idx, var in enumerate(var_l):
            if var in label_d.keys():
                label_l[idx] = label_d[var]
            else:
                label_l[idx] = var
                               
        # Add every single subplot to the figure with a for loop
        for i, var in enumerate(var_l):
            
          # Select the data 
          dcv = dc.loc[var, :].sort_values(by='tau')
          dcv['tau'] = 100*dcv['tau'].copy() # For readibility
          
          # Main frame
          axs[i].plot(dcv.tau, dcv.coeff, lw=3, color='navy')
          axs[i].plot(dcv.tau, dcv.upper_ci, ls='--', color='blue')
          axs[i].plot(dcv.tau, dcv.lower_ci, ls='--', color='blue')

          # Fill in-between
          x = [float(x) for x in dcv.tau.values]
          u = [float(x) for x in dcv.lower_ci.values]
          l = [float(x) for x in dcv.upper_ci.values]

          axs[i].fill_between(x, u, l, facecolor='blue', alpha=0.05)

          # Hline
          axs[i].axhline(y=0, color='black', lw=0.8)

          # Caption
          axs[i].set_title(f'{label_l[i]}', y=1.02)

        # R2 plot
        dr2 = dc.loc['Intercept', :].sort_values(by='tau').copy()
        axs[len(var_l)].plot(100*dr2['tau'], dr2['pseudo_r2'].values,
                             lw=3, color='firebrick')
        axs[len(var_l)].set_title('Pseudo R2', y=1.02)
          
        # Remove extra charts
        for i in range(len(var_l) + 1, len(axs)): 
            axs[i].set_visible(False) # to remove last plot

        ttl = title or (f'Quantile coefficients at horizon {horizon} '
                        f'at {100*self.alpha:.0f}% confidence')    
        fig.suptitle(ttl)
        
        # Return both
        return(fig)


    def term_structure(self, variable, title=None, num_cols=3, 
                       label_d={}, **kwds):
        """ 
        Plot the the coefficients of a single variable across time

        Parameters
        -----------        
        variable: str
          Name of the variable to present the plot

        title: str, default 'Quantile Coefficients and Pseudo R2' 
          Sup title of the plot

        num_cols: int, default 3
          Number of columns, number of rows adjusts automatically

        label_d: dict, default empty
          Label dictionary to replace the subplots caption selectively

        """

        assert variable in self.indvar_l, 'Variable not in regressors list'
        
        # List of horizon
        total_plots = len(self.horizon_l) 

        # Compute the number of rows required
        num_rows = total_plots // num_cols

        if total_plots % num_cols >0:
            num_rows += 1 # Add one row if residuals charts
                
        # Line plot
        dc = self.coeffs.loc[variable, :].copy()

        # Create the main figure
        fig, axs = plt.subplots(nrows=num_rows, ncols=num_cols)

        axs = axs.ravel() # Very helpful !!

        # In case, customize the labels for the plots
        label_l = [None] * len(self.indvar_l)
        
        # Replace the values in list with labels_d
        for idx, var in enumerate(self.indvar_l):
            if var in label_d.keys():
                label_l[idx] = label_d[var]
            else:
                label_l[idx] = var
                               
        # Add every single subplot to the figure with a for loop
        for i, horizon in enumerate(self.horizon_l):
            
          # Select the data 
          dch = dc.loc[dc['horizon']==horizon, :].sort_values(by='tau')
          dch['tau'] = 100*dch['tau'].copy() # For readibility
          
          # Main frame
          axs[i].plot(dch.tau, dch.coeff, lw=3, color='navy')
          axs[i].plot(dch.tau, dch.upper_ci, ls='--', color='blue')
          axs[i].plot(dch.tau, dch.lower_ci, ls='--', color='blue')

          # Fill in-between
          x = [float(x) for x in dch.tau.values]
          u = [float(x) for x in dch.lower_ci.values]
          l = [float(x) for x in dch.upper_ci.values]

          axs[i].fill_between(x, u, l, facecolor='blue', alpha=0.05)

          # Hline
          axs[i].axhline(y=0, color='black', lw=0.8)

          # Caption
          axs[i].set_title(f'Horizon {horizon}', y=1.02)
          
        # Remove extra charts
        for i in range(len(self.horizon_l), len(axs)): 
            axs[i].set_visible(False) # to remove last plot

        ttl = title or (f'Quantile coefficients for {variable} '
                        f'at different horizons, '
                        f'at {100*self.alpha:.0f}% confidence')    
        fig.suptitle(ttl)
        
        # Return both
        return(fig)


    def term_coefficients(self, variable,
                          tau_l=[0.05, 0.25, 0.5, 0.75, 0.95], 
                          title=None, num_cols=3, 
                          label_d={}, **kwds):
        """ 
        Plot the the coefficients of a single variable across time
        With horizon in x-axis

        Parameters
        -----------        
        variable: str
          Name of the variable to present the plot

        title: str, default 'Quantile Coefficients and Pseudo R2' 
          Sup title of the plot

        num_cols: int, default 3
          Number of columns, number of rows adjusts automatically

        label_d: dict, default empty
          Label dictionary to replace the subplots caption selectively

        """

        assert variable in self.indvar_l, 'Variable not in regressors list'

        mv_l = [x for x in tau_l if x not in self.quantile_l]
        assert len(mv_l)==0, f'{mv_l} not in quantile list'
        
        # List of horizon
        total_plots = len(tau_l) 

        # Compute the number of rows required
        num_rows = total_plots // num_cols

        if total_plots % num_cols >0:
            num_rows += 1 # Add one row if residuals charts
                
        # Line plot
        dc = self.coeffs.loc[variable, :].copy()

        # Create the main figure
        fig, axs = plt.subplots(nrows=num_rows, ncols=num_cols)

        axs = axs.ravel() # Very helpful !!

        # In case, customize the labels for the plots
        label_l = [None] * len(self.indvar_l)
        
        # Replace the values in list with labels_d
        for idx, var in enumerate(self.indvar_l):
            if var in label_d.keys():
                label_l[idx] = label_d[var]
            else:
                label_l[idx] = var
                               
        # Add every single subplot to the figure with a for loop
        for i, tau in enumerate(tau_l):
            
          # Select the data 
          dch = dc.loc[dc['tau']==tau, :].sort_values(by='tau')
          
          # Main frame
          axs[i].plot(dch.horizon, dch.coeff, lw=3, color='navy')
          axs[i].plot(dch.horizon, dch.upper_ci, ls='--', color='blue')
          axs[i].plot(dch.horizon, dch.lower_ci, ls='--', color='blue')

          # Fill in-between
          x = [float(x) for x in dch.horizon.values]
          u = [float(x) for x in dch.lower_ci.values]
          l = [float(x) for x in dch.upper_ci.values]

          axs[i].fill_between(x, u, l, facecolor='blue', alpha=0.05)

          # Hline
          axs[i].axhline(y=0, color='black', lw=0.8)

          # Labels
          axs[i].set_xlabel('Horizon', labelpad=10)
          
          # Caption          
          axs[i].set_title(f'Quantile {100*tau:.0f}', y=1.02)
          
        # Remove extra charts
        for i in range(len(self.horizon_l), len(axs)): 
            axs[i].set_visible(False) # to remove last plot

        ttl = title or (f'Term quantile coefficients for {variable} '
                        f'at different quantiles, '
                        f'at {100*self.alpha:.0f}% confidence')    
        fig.suptitle(ttl)
        
        # Return both
        return(fig)

    
    
                
###############################################################################
#%% Plot class for the projection
###############################################################################
class QuantileProjectionPlot(object): # Plot class for the projection class

    """ 
    Plot the output of the different projections
    """

    # Import from QuantileProj class
    def __init__(self, QuantileProjection):
        self.__dict__.update(QuantileProjection.__dict__) # Pass all attributes

                
    # Methods
    def fitted_quantile(self, quantile=0.5, title=None,
                        ylabel='', legendfont=None, legendloc='best'):

        # Prepare the frame
        assert quantile in self.quantile_l, 'quantile not in quantile list'
        dcq = self.cond_quant.loc[self.cond_quant.tau==quantile, :].copy()

        # Plot
        fig, ax = plt.subplots()

        # Line
        ax.plot(dcq['horizon'], dcq['conditional_quantile_mean'],
                label=f'Conditional {100*quantile:.0f} quantile',
                lw=4, color='navy')
        ax.plot(dcq['horizon'], dcq['conditional_quantile_mean_ci_lower'],
                ls='--', label='Lower confidence interval', color='navy')
        ax.plot(dcq['horizon'], dcq['conditional_quantile_mean_ci_upper'],
                ls='--', label='Upper confidence interval', color='navy')

        # Area
        ax.fill_between(dcq['horizon'],
                        dcq['conditional_quantile_mean_ci_lower'],
                        dcq['conditional_quantile_mean_ci_upper'], 
                        alpha=0.15, color='dodgerblue')

        # Layout
        ax.legend(framealpha=0, loc=legendloc, fontsize=legendfont)
        ax.set_xlabel('Horizon', labelpad=20)
        ax.set_ylabel(ylabel, labelpad=20)

        title = title or (f'Conditional {100*quantile:.0f}th quantile '
                          'over forecasting horizon')
        ax.set_title(title, y=1.02)

        return(fig)

    
    def fan_chart(self, ylabel='',
                  title=f'Fan chart at different horizons',
                  legendloc='best', legendfont=None,
                  len_sample=1000, method='linear', len_bs=1000,
                  seed=None):

        # Use the standard sample
        dsample = self.sample(len_sample=len_sample, method=method,
                              len_bs=len_bs, seed=seed)
                
        # Compute the statistics of interest
        tau_l = [0.05, 0.25, 0.5, 0.75, 0.95]

        dss = dsample.groupby(['horizon'])[self.depvar].quantile(tau_l)

        dss = dss.reset_index().copy()
        dss.columns = ['horizon', 'tau', self.depvar]
        dss = dss.set_index(['horizon'])

        # Plot
        fig, ax = plt.subplots()

        # Lines
        ax.plot(dss.loc[dss.tau==0.05, self.depvar],
                label='5%', lw=3, color='red', ls=':')
        ax.plot(dss.loc[dss.tau==0.25, self.depvar],
                label='25%', lw=2, color='black', ls='--')
        ax.plot(dss.loc[dss.tau==0.50, self.depvar],
                label='Median', lw=3, color='black')
        ax.plot(dss.loc[dss.tau==0.75, self.depvar],
                label='75%', lw=2, color='black', ls='--')
        ax.plot(dss.loc[dss.tau==0.95, self.depvar],
                label='95%', lw=3, color='red', ls=':')

        # Area
        ax.fill_between(dss.loc[dss.tau==0.05, self.depvar].index,
                        dss.loc[dss.tau==0.05, self.depvar],
                        dss.loc[dss.tau==0.25, self.depvar],
                        alpha=0.35, color='red')

        ax.fill_between(dss.loc[dss.tau==0.25, self.depvar].index,
                        dss.loc[dss.tau==0.25, self.depvar],
                        dss.loc[dss.tau==0.75, self.depvar],
                        alpha=0.75, color='red')

        ax.fill_between(dss.loc[dss.tau==0.75, self.depvar].index,
                        dss.loc[dss.tau==0.75, self.depvar],
                        dss.loc[dss.tau==0.95, self.depvar],
                        alpha=0.35, color='red')

        # Layout
        ax.legend(framealpha=0, loc=legendloc, ncol=3, fontsize=legendfont)
        ax.set_xlabel('Horizon', labelpad=20)
        ax.set_ylabel(ylabel, labelpad=20)
        ax.set_title(title, y=1.02)

        return(fig)

        



# -*- coding: utf-8 -*-
"""
Distributional GARCH package
Wrapper around the excellent ARCH package by K.Sheppard 
https://arch.readthedocs.io/
Romain Lafarguette 2020, rlafarguette@imf.org
Time-stamp: "2021-09-24 22:48:41 RLafarguette"
"""

###############################################################################
#%% Modules import
###############################################################################
# Standard imports
import os, sys, importlib                               # Operating system
import pandas as pd                                     # Dataframes
import datetime                                         # Dates management
import numpy as np                                      # Numeric package
import scipy                                            # Scientific package
import arch                                             # ARCH/GARCH models

# ARCH package functional imports
from arch.univariate import ARX # Drift model
from arch.univariate import GARCH, EGARCH, EWMAVariance # Volatility models
from arch.univariate import Normal, StudentsT, SkewStudent # Distributions

# Other functional imports
from collections import namedtuple                      # Named tuples
from scipy import stats                                 # Statistical tools 
from statsmodels.distributions.empirical_distribution import ECDF

# Graphics
# NB: for some systems, need to remove matplotlib.use('TkAgg')
import matplotlib                                       # Graphical package
matplotlib.use('TkAgg') # Must be called before importing plt
import matplotlib.pyplot as plt
from matplotlib import cm                               # Colormaps
import matplotlib.colors as mcolors
import seaborn as sns                                   # Graphical package

# Local modules
import joyplot2; importlib.reload(joyplot2)
from joyplot2 import joyplot2

###############################################################################
#%% Ancillary functions
###############################################################################
# Format floating numbers
def formatFloat(num):
  if num % 1 == 0: # Convert into an integer if possible
    return(int(num))
  else: # Keep as it is
    return(num) 

# Flatten nested lists
def flatten(nested_l):
    return([item for sublist in nested_l for item in sublist])

# Find the closest element in a list, return index and value
def closest(value, lst):
    """ Find the closest element in a list and return (index, value) """
    lst = np.asarray(lst) 
    idx = (np.abs(lst - value)).argmin() 
    return(idx, lst[idx])

# Return the next business bay (or later) for any date
def next_business_day(timestamp, horizon=1):
    """ Return the next business bay (or later) for any date """
    # Pay attention: days=0 means next business day (offset 1 by default)
    timestamp = pd.to_datetime(timestamp)
    
    bd = pd.tseries.offsets.BusinessDay(
        offset=datetime.timedelta(days=horizon-1))
    next_day = timestamp + bd
    return(next_day)

# Avoid jumps in plots
# https://stackoverflow.com/questions/1273472/how-to-skip-empty-dates-weekends-in-a-financial-matplotlib-python-graph/2335781
from matplotlib.ticker import FuncFormatter

def equidate_ax(fig, ax, dates, fmt="%Y-%m-%d", label="Date"):
    """
    Sets all relevant parameters for an equidistant date-x-axis.
    Tick Locators are not affected (set automatically)

    Args:
        fig: pyplot.figure instance
        ax: pyplot.axis instance (target axis)
        dates: iterable of datetime.date or datetime.datetime instances
        fmt: Display format of dates
        label: x-axis label
    Returns:
        None

    """    
    N = len(dates)
    def format_date(index, pos):
        index = np.clip(int(index + 0.5), 0, N - 1)
        return dates[index].strftime(fmt)
    ax.xaxis.set_major_formatter(FuncFormatter(format_date))
    ax.set_xlabel(label)
    fig.autofmt_xdate()

###############################################################################
#%% Class: DistGARCH
###############################################################################
class DistGARCH(object):
    """ 
    Estimate a Distribution Model based on GARCH
    With different specifications for drift, volatility and errors distribution
    Based on the excellent ARCH Python package produced by K.Sheppard

    Inputs
    ------
    depvar_str: str
       Name of the dependent variable

    data = pandas dataframe
       Historical data to fit the model on 

    level_str: str
       Name of the dependent variable in level (in case depvar is returns)

    exog_l: list; default None
        List of exogeneous variables for the ARX process

    lags_l: list; default None
        List of lags for the AR(X) process
        
    vol_model: arch volatility model, default GARCH(1,1)
        Volatility model from the arch package. Possible models include:
        ARCH, GARCH, EGARCH, EWMA, etc.
        https://arch.readthedocs.io/en/latest/univariate/volatility.html

    dist_family: arch distribution family, default Normal()
        Family for the distribution of the errors terms. Can include
        Normal(), SkewStudent(), StudentsT()
        https://arch.readthedocs.io/en/latest/univariate/distribution.html

    Output
    ------
    A GARCH class object, wrapping .fit() and .fit().forecast() classes

    Usage:
    dg = DistGARCH(depvar_str, level_str, data)

    """
    __description = "Distributional GARCH"
    __author = "Romain Lafarguette, IMF/MCM, rlafarguette@imf.org"

    # Initializer
    def __init__(self, depvar_str, data, level_str=None, 
                 exog_l=None, lags_l=[1],
                 vol_model=GARCH(1,1),
                 dist_family=Normal(),
                 random_state=42, # Answer to the Ultimate Question of Life 
    ):
        
        # Attributes
        self.depvar = depvar_str
        self.level = level_str
        self.exog_l = exog_l
        self.lags_l = lags_l
        self.rs = np.random.RandomState(random_state) 
        
        # Special attributes
        self.vol_model = vol_model # By default GARCH(1,1)
        
        self.dist_family = dist_family # By default Normal distribution
        self.dist_family._random_state = self.rs
        
        #### Data treatment
        # Aligning all the variables, take care of possible None        
        v_l = [self.depvar]
        if self.level:
            v_l = [self.depvar, self.level]
        if self.exog_l: # Accomodate missing level too
            v_l = v_l + self.exog_l          
        self.df = data.loc[:, v_l].dropna().copy() # align all variables 
                
        # Subselect exogeneous variables if possible
        if self.exog_l:
            self.df_exog = self.df.loc[:, list(self.exog_l)].copy()
        else:
            self.df_exog = None # Keep None
        
        #### Main Models    
        # Mean (drift) model, potentially with exogeneous regressions and lags
        self.mod = ARX(y=self.df[self.depvar], 
                       x=self.df_exog,
                       constant=True, # Fit intercept by defaults
                       lags=self.lags_l)

        # Volatiliy model
        self.mod.volatility = self.vol_model

        # Specify the distribution of the error terms
        self.mod.distribution = self.dist_family 
        
    # Class-methods (methods which returns a class defined below)    
    def fit(self, cov_type='robust', disp='off', update_freq=1):
        return(DistGARCHFit(self, cov_type, disp, update_freq))

    # Public methods
    def plot_description(self, start_date=None,
                         title_returns = 'Historical returns',
                         title_level = 'Historical level',
                         title_density = 'Historical distribution of returns',
                         y_label_returns = 'bps',
                         y_label_level = 'FX rate',
                         xticks_freq=None):
        """ 
        Descriptive plot: returns and level, with distribution

        start_date: str, default None
            To restrict the sample after a given date. Enter as "2015-03-31"

        """

        # Restrict the sample if needed
        if start_date:
            data = self.df.loc[start_date:, :].copy()
        else:
            data = self.df.copy()
            
        # Prepare the plots
        if self.level:
            fig, (ax1, ax2, ax3) = plt.subplots(3,1)

        else:
            fig, (ax2, ax3) = plt.subplots(2,1)

        # First plot: level
        if self.level:
            ax1.plot(data.index, data[self.level])
            ax1.set_title(title_level, y=1.02)
            ax1.set_xlabel('')
            ax1.set_ylabel(y_label_level, labelpad=20)

        # Second plot: Returns
        ax2.plot(data.index, data[self.depvar])
        ax2.set_title(title_returns, y=1.02)
        ax2.set_xlabel('')
        ax2.set_ylabel(y_label_returns, labelpad=20)

        # Manage frequency of xticks & make sure the last one always visible
        if xticks_freq:
            start, end = ax1.get_xlim()
            t_seq = np.append(np.arange(start, end-5, xticks_freq), end)
            ax1.xaxis.set_ticks(t_seq)
            ax2.xaxis.set_ticks(t_seq)
        
        # Third plot: Returns Density
        ax3 = sns.distplot(data[self.depvar])
        ax3.set_title(title_density, y=1.02)
        ax3.set_xlabel(y_label_returns)
        
        # Adjust
        plt.subplots_adjust(hspace=0.5)

        # Exit
        return(fig)
       
###############################################################################
#%% Class: DistGARCHFit
###############################################################################
class DistGARCHFit(object): # Fitted class for the DistGARCH class
    """ 
    Fit a DistGarch model

    General documentation
    https://arch.readthedocs.io/en/latest/univariate/generated/generated/arch.univariate.ARX.fit.html#arch.univariate.ARX.fit

    Inputs
    ------
    cov_type: str
        Type of covariance
    
    disp: str, 'off' or 'on'
        Display the results of the iterations 

    update_freq: integer, default=1
        Frequency of iteration updates
        Output is generated every update_freq iterations    
        Set to 0 to disable iterative output
    
    """

    # Initialization
    def __init__(self, DistGARCH, cov_type, disp, update_freq):

        self.__dict__.update(DistGARCH.__dict__) # Import all attributes

        # Fit the model (just take from the ARCH model)    
        self.res = self.mod.fit(
            cov_type=cov_type, 
            disp=disp, 
            update_freq=update_freq)

        # Print the summary at each fit
        print(self.res.summary())

        # Erros Distribution parameters (depends on the distribution)
        # Note that the distributions are NORMALIZED
        if self.mod.distribution.name=='Normal':
            self.dist_params = np.array([0,1]) # By default
            self.scipy_dist = scipy.stats.norm
            
        elif self.mod.distribution.name=="Standardized Student's t":
            self.dist_params = self.res.params['nu']
            self.scipy_dist = scipy.stats.t
            
        elif self.mod.distribution.name=="Standardized Skew Student's t":
            self.dist_params = self.res.params[['nu', 'lambda']]
            self.scipy_dist = scipy.stats.nct
            
        elif self.mod.distribution.name=="Generalized Error Distribution":
            self.dist_params = self.res.params['nu']
            self.scipy_dist = scipy.stats.gennorm
                                    
        else:
            raise ValueError('Distribution name mis-specified')


        # Compute the historical and in-sample volatility and mean
        self.hist_avg = self.df[self.depvar].mean()
        self.hist_vol = self.df[self.depvar].std()
        self.in_cond_vol = self.res.conditional_volatility
               
    # Class-methods (methods which returns a class defined below)    
    def forecast(self, start_date, horizon=1, fmethod='analytic',
                 sample_size=10000):
        return(DistGARCHForecast(self, start_date, horizon, fmethod,
                                 sample_size))

    # Public methods: Shocks Simulation
    def shock_simulate(self, nobs=100, burn=0, mean_mult=1, vol_mult=1):
        """ 
        Simulate the model based on estimated parameters 

        nobs: integer, default 100
            Number of observations of the simulated frame

        burn: integer, default 0
            Number of initial observations to "burn". Put at 0 to study shocks

        mean_mult: float, default 1
            Multiplier on hist average to adjust the initial shock

        vol_mult: float, default 1
            Multiplier on hist volatility to increase the initial shock

        """
                
        # Extract the parameters without exogeneous variables
        # Exogeneous variables can not be forecasted...
        if self.exog_l:
            spar_index = [x for x in self.res.params.index
                          if x not in self.exog_l]
        else:
            spar_index = self.res.params.index
    
        vparams = self.res.params[spar_index]

        # Simulate the model
        # Linear volatility models (in x)
        # ARCH => initial_value_vol=np.power(x*hist_vol,2) (variance)
        # GARCH => initial_value_vol=np.power(x*hist_vol,2) (variance)
        # EWMA => initial_value_vol=np.power(x*hist_vol,2) (variance)

        # Non-linear volatility models (linear if transform x...)
        # EGARCH => initial_value_vol=np.log((x*hist_vol)**2) (log of variance)

        # For processes linear in variance (not vol)
        if self.mod.volatility.name in ['ARCH', 'GARCH', 'GJR-GARCH',
                                        'EWMA/RiskMetrics',
                                        'RiskMetrics2006']:
            init_vol = np.power(vol_mult*self.hist_vol,2)
        elif self.mod.volatility.name in ['EGARCH']: # Exponential vol
            init_vol = np.log((vol_mult*self.hist_vol)**2)
        else:
            raise ValueError((f'Volatility model {self.mod.volatility.name}'
                             ' not incorporated yet'))

        # Run the simulation
        dsim = self.res.model.simulate(vparams, nobs=nobs, burn=burn,
                                       initial_value= mean_mult*self.hist_avg,
                                       initial_value_vol= init_vol)

        # Provide normalized volatility
        dsim['norm_volatility'] = dsim['volatility']/self.hist_vol
        return(dsim)

    # Public methods: summary table
    def summary_table(self, model_name='GARCH model', var_d=None,
                      round_param=2,
                      print_pval=True):
        """
        Inputs
        ------

        model_name=str, default "GARCH"
            Name of the model which will appear in the table column

        var_d: dict, default None
            Dictionary for renaming the variables as index

        round_param: int, default 3
            Rounding parameter for the model coefficients

        """

        
        
        # Preparation of the table
        st = 'Significance *10%, **5%, ***1%'
        if print_pval==True:
            var_l = list(self.res.params.index) + ['R2', 'R2 adjusted',
                                                   'Number of observations', 
                                                   'Pvalue in parenthesis',
                                                   st]
        else:
            var_l = list(self.res.params.index) + ['R2', 'R2 adjusted',
                                                   'Number of observations', 
                                                   st]


        summary_table = pd.DataFrame(index=var_l, columns=[model_name])

        # Populate the table
        for variable in self.res.params.index:
            param = self.res.params[variable]
            pval = self.res.pvalues[variable]

            if (pval<=0.1) & (pval>0.05):
                stars='*'
            elif (pval<=0.05) & (pval>0.01):
                stars='**'
            elif pval<=0.01:
                stars='***'
            else:
                stars=''

            if print_pval==True:    
                txt = (f'{str(round(param, round_param)) + str(stars)}'
                       f'({str(round(pval, round_param))})')
            elif print_pval==False:
                txt = (f"{str(round(param, round_param)) + str(stars)}")
            else:
                raise ValueError('pval parameter misspecified')
            
            summary_table.loc[variable, model_name] = txt

        # Add the R2
        nobs = self.res.nobs
        r2 = f'{round(100*self.res.rsquared,1)} %'
        r2_adj = f'{round(100*self.res.rsquared_adj,1)} %'
        summary_table.loc['R2', model_name] = r2
        summary_table.loc['R2 adjusted', model_name] = r2_adj

        # Add extra information               
        summary_table.loc['Number of observations', model_name] = nobs
                       
        # Customization of variables names, if needed
        rename_d = {'Const':'Intercept',
                    'alpha[1]':'Alpha',
                    'beta[1]':'Beta',
                    'gamma[1]':'Gamma',
                    'omega':'Omega', 
                    'nu':'Nu', 
                    'lambda':'Lambda'}

        if var_d:
            rename_d.update(var_d)

        summary_table = summary_table.rename(rename_d, axis='index')
        summary_table_nna = summary_table.fillna('').copy()

        return(summary_table_nna)
    
    # Public methods: Plots
    def summary_fit_plot(self):
        """ Summary fit plot: residuals and conditional volatility """
        # Prepare the plot
        fig = plt.figure()        
        self.res.plot()
        return(fig)


    # In-sample conditional volatility
    def plot_in_cond_vol(self,
                         start_date=None, 
                         title='In sample conditional volatility',
                         ylabel='Conditional volatility',
                         xticks_freq=None):
        
        """ 
        Plot the (estimated) in-sample conditional volatility 

        start_date: str, default None
            Start date to restrict the sample, str should be '2015-03-31'

        xticks_freq: int or None, default None
            E.g. take one ticks over 5 (None to use plt default)

        """

        # Conditional volatility data
        cv = self.res.conditional_volatility.dropna().copy()

        if start_date:
            cv = cv.loc[start_date:].copy()
                        
        fig, ax1 = plt.subplots(1,1)

        ax1.plot(cv.index, cv, lw=3)
        ax1.set_title(title, y=1.02)
        ax1.set_xlabel('')
        ax1.set_ylabel(ylabel, labelpad=20)
        
        # Manage frequency of xticks & make sure the last one always visible
        if xticks_freq:
            start, end = ax1.get_xlim()
            t_seq = np.append(np.arange(start, end-5, xticks_freq), end)
            ax1.xaxis.set_ticks(t_seq)

        ax1.tick_params(axis='x', rotation=45)    
        return(fig)
    
           
    # Plot the relationship between past innovations and volatility
    def plot_shocks_vol(self,
                        vol_thresholds_l=[1.5, 3],
                        xlabel='Shocks (t-1)',
                        ylabel='Normalized conditional volatility (t, std)', 
                        title='Shocks & conditional volatility',
                        subtitle='With polynomial fit & intervention thresholds',
                        x_range=None, 
                        range_share=0.05):

        """ 
        Scatter Plot of Conditional Volatility & Innovations 
        With polynomial and tangent

        vol_thresholds_l: list of float
            Threshold for volatility. Should be 1.5 for 50% more than min one
        
        share_range: float, between 0 and 1, default 0.1
            Length of the tangent lines, expressed as share of x-data range

        """
                      
        #### Data work
        ds = pd.merge(self.res.resid, self.in_cond_vol,
                      left_index=True, right_index=True)
        ds.columns = ['shocks', 'cond_vol']

        ds = ds.dropna().copy()        
        
        # Normalize volatility by a constant to improve the charts
        # Look at the volatility with the constant of a quadratic fit
        qparams = np.polyfit(ds['shocks'], ds['cond_vol'], 2)
        constant = qparams[2] # "c" in : ax2 + bx + c

        # Look at the volatility in normal times, 
        #normalization = np.percentile(ds['cond_vol'], 50)
        ds['norm_cond_vol'] = ds['cond_vol']/constant

        # Data
        x_data = ds['shocks'].values
        y_data = ds['norm_cond_vol'].values

        # Polynomial fit
        params = np.polyfit(x_data, y_data, 2)

        if x_range:
            x_support = np.linspace(x_range[0], x_range[1], 1000)
        else:
            x_support = np.linspace(min(x_data), max(x_data), 1000)
            
        y_fit = np.polyval(params, x_support)
           
        #### Plots        
        # Prepare the plot
        fig, ax = plt.subplots(1,1)

        # Draw original data as a scatter plot
        ax.scatter(x_data, y_data, color='blue')
        ax.plot(x_support, y_fit, color='green', lw=3)

        t_idx_l = list() # Container to store threshold lists
        for thresh in vol_thresholds_l:
            ax.axhline(y=thresh, linestyle='--', color='darkred')
          
            # Point where the intersection occurs
            inter_idx = np.argwhere(np.diff(np.sign(thresh - y_fit))).flatten()
            for idx in inter_idx:
                t_idx_l.append(idx)
                ax.vlines(x=x_support[idx], ymin=0, ymax=y_fit[idx],
                          color='darkred', linestyle='--')
                         
        ax.scatter(x_support[t_idx_l], y_fit[t_idx_l], s=100, c='darkred')  

        # Arrange the y-ticks        
        ax.yaxis.set_ticks(np.arange(0, max(y_fit) + 0.5, 0.5))
        
        # Remove the standard x-ticks
        ax.set_xticks([])
        new_ticks_l = sorted([int(x) for x in x_support[t_idx_l]])
        ax.set_xticks(new_ticks_l + [0]) # Add new ticks

        extra_idx_l = [new_ticks_l.index(x) for x in new_ticks_l]
        for idx in extra_idx_l: 
            ax.get_xticklabels()[idx].set_color("darkred")
            #ax.get_xticklabels()[idx].set_fontproperties('bold')
        
        # Some customization
        ax.set_xlabel(xlabel, labelpad=20) # X axis data label
        ax.set_ylabel(ylabel, labelpad=20) # Y axis data label
        
        x_range = max(x_support) - min(x_support)
        plt.xlim(min(x_support) - range_share*x_range,
                 max(x_support) + range_share*x_range)
 
        plt.ylim(0, max(y_fit))

        
        # Title
        plt.title(f'{title} \n {subtitle}', y=1.02)        

        plt.show()

        return(None)
       
###############################################################################
#%% Class: DistGARCHForecast
###############################################################################
class DistGARCHForecast(object): # Forecast class for the DistGARCHFit class
    """ 
    Forecast from a DistGarchFit model

    General documentation on VaR forecasting
    https://arch.readthedocs.io/en/latest/univariate/univariate_volatility_forecasting.html#Value-at-Risk-Forecasting
    
    Inputs
    ------
    start_date: str
        Starting date for the forecasts. Example: '2020-03-31'

    horizon: int
        Forecasting horizon
        https://arch.readthedocs.io/en/latest/univariate/univariate_volatility_forecasting.html#Fixed-Window-Forecasting

    fmethod: str
        Simulation methods, among 'analytic', 'simulation', 'bootstrap'
        Not all methods are available for all horizons.
        https://arch.readthedocs.io/en/latest/univariate/forecasting.html#simulation-forecasts

    sample_size: int
        Size of the sample for distribution sampling

    """
    # Initialization
    def __init__(self, DistGARCHFit, start_date, horizon, fmethod,
                 sample_size):

        self.__dict__.update(DistGARCHFit.__dict__) # Import all attributes 
        
        self.start_date = start_date
        self.horizon = horizon
        self.fmethod = fmethod
        self.sample_size = sample_size

        # Construct the forecasted matrix of exogeneous regressors
        self.exog_fcast_d = dict() # Construct the dictionary of regressors

        if len(self.exog_l)>0:
            for var in self.exog_l:
                #val = self.df_exog.loc[self.start_date, [var]].values
                #self.exog_fcast_d[var] = np.full((len(self.df_exog), 1), val)
                self.exog_fcast_d[var] = self.df_exog.loc[:, [var]].values
                
        elif self.exog_l == None: # No exogeneous regressors
                self.exog_fcast_d = None
        else:
            raise ValueError('List of exogeneous regressors misspecified')
                
        # Run the forecasts from the arch package
        self.forecasts = self.res.forecast(horizon=self.horizon,
                                           x=self.exog_fcast_d, 
                                           start=self.start_date, 
                                           method=self.fmethod,
                                           reindex=True)
                                                        
        # Extract the forecasted conditional mean and variance
        self.cond_mean = self.forecasts.mean[start_date:]
        self.cond_var = self.forecasts.variance[start_date:]
        
        dfor = pd.merge(self.cond_mean, self.cond_var,
                        left_index=True, right_index=True)
        self.dfor = pd.DataFrame(dfor.values, index=self.cond_var.index,
                                 columns=['cond_mean', 'cond_var'])
        
        # Compute the normalized true values (for the PIT test)
        avl_idx_l = [x for x in self.df.index if x in self.dfor.index]
        self.dfor.insert(0, 'true_val',
                         self.df.loc[avl_idx_l, self.depvar])
        self.dfor['cond_vol'] = np.sqrt(self.dfor['cond_var'])

        # Because the distribution are on the normalized values
        self.dfor['norm_true_val'] = ((self.dfor['true_val']
                                      - self.dfor['cond_mean'])
                                      /self.dfor['cond_vol'])
        
        #### CDF and sampling estimation
        # Important: the standardized distribution does NOT change over time
        # Parameters are fixed, The conditional mean and volatility change
        q_l = [0.01] + [round(x,3) for x in np.arange(0.025,1,0.025)] + [0.99]
        self.cond_quant_labels_l = [f"cond_quant_{q:g}" for q in q_l]
        
        # Define functions associated with conditional distribution 
        if self.mod.distribution.name=='Normal': # No parameters
            self.cond_ppf = lambda q: self.mod.distribution.ppf(q)
            self.cond_cdf = lambda v: self.mod.distribution.cdf(v)
            
        else:
            self.cond_ppf = lambda q: self.mod.distribution.ppf(q,
                                                            self.dist_params)
            self.cond_cdf = lambda v: self.mod.distribution.cdf(v,
                                                            self.dist_params)
            
        err_cond_quant = self.cond_ppf(q_l)
        self.dfor['pit'] = self.cond_cdf(self.dfor['norm_true_val'])
        self.dfor['norm_true_val_cdf'] = self.dfor['pit'].copy() # Simple name
        
        self.mod.distribution._random_state = self.rs
        err_sampler = self.mod.distribution.simulate(self.dist_params)
        
        # PAY ATTENTION THAT THE DISTRIBUTIONS ARE STANDARDIZED
        # TO RECOVER THE SERIES, NEED TO REINFLATE WITH MEAN AND VAR
        
        # Summarize the conditional quantiles in a dataframe
        # xbar = (x-mu)/std => x = mu + xbar*std
        # mean + sqrt(cond_var)*the quantiles of the errors terms
        for var in self.cond_quant_labels_l: self.dfor[var] = np.nan
        self.dfor[self.cond_quant_labels_l] = (self.dfor[['cond_mean']].values
                                            + (self.dfor[['cond_vol']].values
                                                  * err_cond_quant[None, :]))
                      
        # Sampling the error terms and deriving the values for y in a dataframe
        err_sample = err_sampler(self.sample_size)
        self.sample = self.cond_mean.values +(np.sqrt(self.cond_var).values
                                                  * err_sample)
        self.sample = pd.DataFrame(self.sample,
                                   columns=range(self.sample_size),
                                   index=self.dfor.index)

    # Public methods
    def dist_fit(self, fdate):
        """
        Return the scipy random variable for a given date
        """
        
        # Take a sample at a given date
        sample = self.sample.dropna()

        if fdate:
            assert fdate in sample.index, "Date not in data sample"
            ssample = sample.loc[fdate, :].values
        else: # Take the last date available
            ssample = sample.tail(1).values
            fdate = sample.tail(1).index[0].strftime('%Y-%m-%d')

        # Fit the distribution
        params_fit = self.scipy_dist.fit(ssample)
        rv = self.scipy_dist(*params_fit) # Frozen random variate

        # Return the scipy random variable
        return(rv)
        
    def fixed_thresholds_FXI(self, thresholds_t):
        """ 
        Return a dataframe with fixed thresholds 

        Inputs
        ------

        thresholds_t: tuple
            Tuple of intervention threshold, one negative, one positive
                  
        """

        assert isinstance(thresholds_t, tuple), 'Thresholds should be tuple'
        assert thresholds_t[0] < 0, 'First threshold should be negative'
        assert thresholds_t[1] > 0, 'Second threshold should be negative'
        
        # Prepare the frame
        dv = pd.merge(self.dfor, self.df[[self.depvar, self.level]],
                      left_index=True, right_index=True)

        # Determine which values are below or above thresholds

        c_below = (dv['FX log returns'] <= thresholds_t[0])
        c_above = (dv['FX log returns'] >= thresholds_t[1])

        # FX interventions
        dv['FXI'] = 'No'                
        dv.loc[c_below, 'FXI'] = 'Below'
        dv.loc[c_above, 'FXI'] = 'Above'

        # Logreturns where the FXI occur
        dv[f'Logret Below'] = np.nan
        dv.loc[c_below, f'Logret Below'] = dv.loc[c_below, 'FX log returns']
        dv[f'Logret Above'] = np.nan
        dv.loc[c_above, f'Logret Above'] = dv.loc[c_above, 'FX log returns']  
        
        # Level where the FXI occur
        dv[f'Level Below'] = np.nan
        dv.loc[c_below, f'Level Below'] = dv.loc[c_below, self.level]
        dv[f'Level Above'] = np.nan
        dv.loc[c_above, f'Level Above'] = dv.loc[c_above, self.level]
    
        return(dv)


    def VaR_FXI(self, qv_l=[0.025, 0.975]):
        """ 
        Return a dataframe with VaR thresholds

        Inputs
        ------

        qv_l: List
            List of quantiles
                  
        """

        qv_l = sorted(qv_l)
        qv_labels_l = [f'cond_quant_{x:g}' for x in qv_l]
        
        # Prepare the frame
        dv = pd.merge(self.dfor, self.df[[self.depvar, self.level]],
                      left_index=True, right_index=True)

        # Single flag
        dv['FXI'] = np.nan
        
        # Create the exceedance markers, based on quantiles list
        dv['Below'] = np.nan
        c_below = (dv[self.depvar] <= dv[f'{qv_labels_l[0]}'])
        dv.loc[c_below, 'Below'] = dv.loc[c_below, self.depvar]
        dv.loc[c_below, 'FXI'] = 'Below'

        dv['Above'] = np.nan
        c_above = (dv[self.depvar] >= dv[f'{qv_labels_l[1]}'])
        dv.loc[c_above, 'Above'] = dv.loc[c_above, self.depvar]
        dv.loc[c_above, 'FXI'] = 'Above'
                        
        if self.level: # If level, add it
            dv[f'Level Below'] = np.nan
            dv.loc[c_below, f'Level Below'] = dv.loc[c_below, self.level]

            dv[f'Level Above'] = np.nan
            dv.loc[c_above, f'Level Above'] = dv.loc[c_above, self.level]

        return(dv)

    # Public plot methods
    # Forecasted conditional mean and volatility
    def plot_out_cond_mean_vol(self, ylabel='',
                               title_mean='Out-of-sample conditional mean',
                               title_vol='Out-of-sample conditional volatility', 
                               ylabel_mean='',
                               ylabel_vol='',
                               xticks_freq=None, 
                               hspace=0.3):
        """ 
        Plot the conditional mean and volatility over forecasted periods
        """

        # Prepare the plot and fix the parameters                
        fig, (ax1, ax2) = plt.subplots(2, 1)

        # Conditional mean
        ax1.plot(self.dfor.index, self.dfor['cond_mean'])
        ax1.set_title(title_mean, y=1.02)
        ax1.set_xlabel('')
        ax1.set_ylabel(ylabel_mean, labelpad=20)

        if xticks_freq:
            start, end = ax1.get_xlim()
            t_seq = np.append(np.arange(start, end-5, xticks_freq), end)
            ax1.xaxis.set_ticks(t_seq)

        # Conditional volatility
        ax2.plot(self.dfor.index, self.dfor['cond_vol'])
        ax2.set_title(title_vol, y=1.02)
        ax2.set_xlabel('')
        ax2.set_ylabel(ylabel_vol, labelpad=20)

        if xticks_freq:
            start, end = ax2.get_xlim()
            t_seq = np.append(np.arange(start, end-5, xticks_freq), end)
            ax2.xaxis.set_ticks(t_seq)
        
        # Adjust the space between plots
        plt.subplots_adjust(hspace=hspace)
        
        return(fig)

    # Conditional Quantiles
    def plot_out_condquant(self, quantile=0.05,
                           ylabel='',
                           title='Out-of-sample conditional VaR at 5%'):
        """ 
        Plot the conditional value at risk (quantile) over forecasted periods

        clip: tuple or None, default None
            Tuple to limit the plots

        """
        # Prepare the plot        
        fig = plt.figure()        
        self.dfor[f'cond_quant_{quantile:g}'].plot(lw=3, legend=None)
        plt.xlabel('')
        plt.ylabel(ylabel, labelpad=20)
        plt.title(title, y=1.01)
        return(fig)

    # Multiple distributions plot (Joyplot)
    def joyplot_out(self,
                    title='Conditional forecasted density out-of-sample',
                    label_drop=3,
                    colormap=cm.autumn_r,
                    xlabel='',
                    xlimits_t=None):
        """ 
        Plot a forecasted joyplot (multiple densities plotted over time) 

        label_drop: integer, default 3
            Keep one label every x of them, avoiding crowded chart

        colormap: matplotlib colormap. Default: yellow/red scale
            To have a unique color (e.g. all blue), choose None

        xlimits_t: tuple or None, default None
            Fix the limits for the xscale

        """
        
        # Arrange the sample frame        
        dsam = self.sample.dropna().transpose().copy()
        dates_l = dsam.columns 
        dsam['id'] = dsam.index
        dlsam = pd.melt(dsam, value_vars=dates_l,
                        id_vars=['id'])
        dlsam = dlsam.rename(columns={'variable':'date'})
        dlsam['fdate'] = dlsam['date'].dt.strftime('%Y-%m-%d') 

        # Colors based on the volatility of each sample (can do on median too)
        vs = pd.DataFrame(dlsam.groupby(['date'])['value'].std())
        vs['date'] = pd.to_datetime(vs.index)
        vs['fdate'] = vs['date'].dt.strftime('%Y-%m-%d') 
        vs = vs.rename(columns= {'value': 'volatility'})

        colorparams = vs['volatility'] # Parameter of the color
        colormap = cm.autumn_r
        normalize = mcolors.Normalize(vmin=np.min(colorparams),
                                      vmax=np.max(colorparams))
        vs['color'] = [mcolors.to_rgba(colormap(normalize(x)))
                       for x in vs['volatility']]

        # Need to map volatility and color
        dall = pd.merge(dlsam, vs, left_on=['fdate'], right_on=['fdate'])
        
        # Arrange the labels, take one every three
        labels_l = list()
        for idx,val in enumerate(sorted(set(dall['fdate']))):
            if idx%label_drop==0:
                labels_l.append(val)
            else:
                labels_l.append(None)

        # Avoid extreme values        
        if isinstance(xlimits_t, tuple):
            dall = dall.loc[dall['value'] >= xlimits_t[0], :].copy()
            dall = dall.loc[dall['value'] <= xlimits_t[1], :].copy()


        # Prepare the plot
        fig, axes = joyplot2(dall, by="fdate", column="value",
                             range_style='own',
                             labels=labels_l, figsize=(12,10),
                             grid="y", linewidth=1, legend=False, 
                             fade=True,
                             color=list(vs['color']))
    
        plt.xlabel(xlabel)
        plt.title(title, y=1.02) # Can not be on left
            
        return(fig)
    
    # Plot the fxi rule as a pdf
    def plot_pdf_rule(self, fdate=None, q_low=0.05, q_high=0.95,
                      title=None, 
                      xlabel='bps', ylabel='density',
                      sample_lim=0.1, ax=None):
        """ 
        Summary chart of FX intervention rule

        fdate: str or None, default None
            Date to condition the data for the forecast. 
            By default, last date available

        q_low, q_high: float, between 0 and 1
            The quantiles of interest, on the left and right

        I should recode it to use directly the pdf instead of the sample...
        Not difficult, but I am tired now and I need to finish in a hurry


        """

        # Take a sample at a given date
        sample = self.sample.dropna()

        if fdate:
            assert fdate in sample.index, "Date not in data sample"
            ssample = sample.loc[fdate, :].values
        else: # Take the last date available
            ssample = sample.tail(1).values
            fdate = sample.tail(1).index[0].strftime('%Y-%m-%d')

        # Fit the distribution
        # I am lazy, the arch dist have no pdf readily done so I fit a sample..
        params_fit = self.scipy_dist.fit(ssample)
        rv = self.scipy_dist(*params_fit) # Frozen random variate
        support = np.linspace(np.percentile(ssample, sample_lim),
                              np.percentile(ssample, 100-sample_lim), 1000)

        # Compute the pdf
        pdf = rv.pdf(support)
        #pdf = self.scipy_dist.pdf(support)
        
        # Compute the quantiles to determine the intervention region
        qval_low = rv.ppf(q_low)
        qval_pdf_low = rv.pdf(qval_low)
        qval_high = rv.ppf(q_high)
        qval_pdf_high = rv.pdf(qval_high)

        # Compute the mode
        x_mode = support[list(pdf).index(max(pdf))]
        y_mode = rv.pdf(x_mode)

        
        # Prepare the plot        
        fig, ax = plt.subplots(1,1)
        ax.plot(support, pdf, lw=3)

        # Intervention regions
        ax.fill_between(support, 0, pdf, where=support<=qval_low, color='red',
                        label='Intervention region')
        ax.fill_between(support, 0, pdf, where=support>qval_high, color='red')

        #### Add text and lines about VaR and Mode
        # Low quantile
        ax.text(qval_low, qval_pdf_low , f'VaR {100*q_low}%', color='darkred',
                horizontalalignment='right', verticalalignment='bottom')
        ax.text(0.99*qval_low, 0, '{:.1f}'.format(qval_low),
                horizontalalignment='left', color='darkred',
                verticalalignment='top')
        ax.vlines(qval_low, 0, qval_pdf_low, linestyle='--')

        # High quantile
        ax.text(qval_high, qval_pdf_high , f'VaR {100*q_high}%',
                color='darkred',
                horizontalalignment='left', verticalalignment='bottom')
        ax.text(0.99*qval_high, 0, '{:.1f}'.format(qval_high),
                horizontalalignment='right', color='darkred',
                verticalalignment='top')
        ax.vlines(qval_high, 0, qval_pdf_high, linestyle='--')

        # Mode
        ax.text(x_mode, y_mode , f'Mode', color='darkred',
                horizontalalignment='center', verticalalignment='bottom')
        ax.text(x_mode, 0, '{:.1f}'.format(x_mode),
                horizontalalignment='center', color='darkred',
                verticalalignment='top')
        ax.vlines(x_mode, 1.1*min(pdf), y_mode, linestyle='--')

        
        ax.set_xlabel(xlabel, labelpad=20) # X axis data label
        ax.set_ylabel(ylabel, labelpad=20) # Y axis data label

        # Final customizations
        plt.xlim(min(support), max(support))
        plt.legend()

        if title:
            plt.title(title, y=1.02)
        else:
            ttl = (f'Conditional density and intervention rule'
                   f' based on {fdate} information')
            plt.title(ttl, y=1.02)

        if ax:
            return(ax)
        else:
            return(fig)


    # VaR exceedance
    def plot_var_exceedance(self, qv_l=[0.025, 0.975], 
                            title_1='Log Returns and VaR Exceedance', 
                            title_2='Level',
                            swap_color=None, 
                            y1_label='',
                            y2_label='',
                            size=100):
        
        """ 
        Plot the VaR Exceedance 
        On the dependent variable, also possible to indicate level

        qv_l: list of two quantiles, 
              List of the upper and below quantiles
 

        """

        qv_l = sorted(qv_l)
        qv_labels_l = [f'cond_quant_{x:g}' for x in qv_l]

        
        # Prepare the frame
        dv = self.VaR_FXI(qv_l)
        
        # Subselect the frame and plot it
        dvs = dv.loc[self.start_date:, :].copy()

        # Prepare the plot
        if self.level:
            fig, (ax1, ax2) = plt.subplots(2,1)
        else:
            fig, ax1 = plt.subplots(1,1)

        x = np.arange(len(dvs.index))

        if swap_color:
            ctop = 'green'
            cdown = 'red'
        else:
            ctop = 'red'
            cdown = 'green'
        
        # First plot
        ax1.plot(x, dvs[self.depvar], lw=3)
        ax1.scatter(x, dvs['Below'], alpha=0.8, c=ctop, marker='D', s=size,
                    label=f'Below VaR {qv_labels_l[0]}%')
        ax1.scatter(x, dvs['Above'], alpha=0.8, c=cdown, marker='o', s=size,
                    label=f'Above VaR {qv_labels_l[1]}%')
        equidate_ax(fig, ax1, dvs.index) # Adjust the dates
        ax1.set_title(title_1, y=1.02)
        ax1.set_xlabel('')
        ax1.set_ylabel(y1_label, labelpad=20)

        if self.level:
            # Second plot
            ax2.plot(x, dvs[self.level], lw=3)
            ax2.scatter(x, dvs['Level Below'], alpha=0.8, s=size,
                        c=ctop, marker='D', label='Level Below')
            ax2.scatter(x, dvs['Level Above'], alpha=0.8, s=size,
                        c=cdown, marker='o', label='Level Above')
            equidate_ax(fig, ax2, dvs.index) # Adjust the dates
            ax2.set_title(title_2, y=1.02)
            ax2.set_xlabel('')
            ax2.set_ylabel(y2_label, labelpad=20)

        plt.subplots_adjust(hspace=0.3)
        return(fig)


    def plot_fixed_exceedance(self, thresholds_t=(-200, 200), 
                              title_1='Log Returns and Fixed Thresholds '
                              'Exceedance', 
                              title_2='FX Level',
                              swap_color=None, 
                              y1_label='',
                              y2_label='',
                              size=100, hspace=0.3):
        
        """ 
        Plot the Fixed Thresholds Exceedance 
        On the dependent variable, also possible to indicate level

        thresholds_l: tuple of fixed thresholds for FXI, 
              
        """ 

        # Generate the data frame
        dv = self.fixed_thresholds_FXI(thresholds_t)

        # Subselect the frame and plot it
        dvs = dv.loc[self.start_date:, :].copy()

        # Prepare the plot
        fig, (ax1, ax2) = plt.subplots(2,1)

        x = np.arange(len(dvs.index))

        # Color management
        if swap_color:
            ctop = 'green'
            cdown = 'red'
        else:
            ctop = 'red'
            cdown = 'green'

        # First plot
        ax1.plot(x, dvs[self.depvar], lw=3)
        ax1.scatter(x, dvs['Logret Below'], alpha=0.8, c=ctop,
                    marker='D', s=size)
        ax1.scatter(x, dvs['Logret Above'], alpha=0.8, c=cdown,
                    marker='o', s=size)
        equidate_ax(fig, ax1, dvs.index) # Adjust the dates
        ax1.set_title(title_1, y=1.02)
        ax1.set_xlabel('')
        ax1.set_ylabel(y1_label, labelpad=20)


        # Second plot
        ax2.plot(x, dvs[self.level], lw=3)
        ax2.scatter(x, dvs['Level Below'], alpha=0.8, s=size,
                    c=ctop, marker='D', label='Level Below')
        ax2.scatter(x, dvs['Level Above'], alpha=0.8, s=size,
                    c=cdown, marker='o', label='Level Above')
        equidate_ax(fig, ax2, dvs.index) # Adjust the dates
        ax2.set_title(title_2, y=1.02)
        ax2.set_xlabel('')
        ax2.set_ylabel(y2_label, labelpad=20)

        plt.subplots_adjust(hspace=hspace)

        plt.show()


    # Plot of the conditional cdf
    def plot_conditional_cdf(self, q_low=0.05, q_high=0.95,
                             title=('Conditional cumulative distribution'
                                    ' function and intervention thresholds'),
                             ylabel='quantile',
                             thresholds_t=None, 
                             swap_color=None,
                             size=100, 
                             xticks_freq=None):

        """ 
        Plot the conditional cdf over time and the intervention thresholds 
        I should recode it to use directly the cdf instead of the sample...
        Not difficult, but I am tired now and I need to finish in a hurry

        xticks_freq: None or int, default None
            Frequency of xticks: e.g., 2 is 1 out of 2. None for plt default

        """

        # Data work
        dcq = self.dfor[self.cond_quant_labels_l].dropna().copy()
        dcq.columns = [x.replace('cond_quant_', '') for x in dcq.columns]

        dates_l = dcq.index[:-1] # Don't take the last one

        dclose = pd.DataFrame(index=dates_l,
                              columns=['realization','cond_quant'])

        # Find the closest quantile in the list
        for fdate in dates_l:
            realization = self.df.loc[fdate, self.depvar]
            cond_quant_l = dcq.loc[fdate, :]
            closest_quantile_idx = closest(realization, cond_quant_l)[0]
            closest_quantile = cond_quant_l.index[closest_quantile_idx]
            dclose.loc[fdate, :] = [realization, closest_quantile]

        dclose['Quantile'] = 100*(dclose['cond_quant'].astype(float))

        
        if thresholds_t:
            # Add the fixed thresholds
            dv = self.fixed_thresholds_FXI(thresholds_t)
            dclose = dclose.merge(dv[['FXI']],
                                  left_index=True, right_index=True)

            da = dclose.loc[dclose['FXI']=='Above', :].copy()
            db = dclose.loc[dclose['FXI']=='Below', :].copy()
            
        else:
            dv = self.VaR_FXI(qv_l=[q_low, q_high])
            dclose = dclose.merge(dv[['FXI']],
                                  left_index=True, right_index=True)
            da = dclose.loc[dclose['FXI']=='Above', :].copy()
            db = dclose.loc[dclose['FXI']=='Below', :].copy()
            
        # Prepare the plot
        fig, ax = plt.subplots()
        ax.plot(dclose.index, dclose['Quantile'], lw=3)

        # Scatter lines with fixed thresholds
        if swap_color:
            ctop = 'green'
            cdown = 'red'
        else:
            ctop = 'red'
            cdown = 'green'


        ax.scatter(da.index, da['Quantile'], alpha=0.8, c=ctop,
                   marker='D', s=size)
        ax.scatter(db.index, db['Quantile'], alpha=0.8, c=cdown,
                   marker='o', s=size)
            

        # Horizontal lines
        ax.axhline(y=100*q_low, color='red', linestyle='--', lw=2)
        ax.axhline(y=100*q_high, color='red', linestyle='--', lw=2)
        ax.axhline(y=25, color='blue', linestyle='-', lw=1)
        ax.axhline(y=50, color='black', linestyle='-', lw=1)
        ax.axhline(y=75, color='blue', linestyle='-', lw=1)

        # Add the ticks, if needed
        new_t_l = [100*q_low, 25, 50, 75, 100*q_high]
        #new_ticks_l = sorted(list(ax.get_yticks()) + new_t_l)
        new_ticks_l = new_t_l
        extra_idx_l = [new_ticks_l.index(x) for x in new_t_l]
        ax.set_yticks(new_ticks_l) # Add new ticks

        for idx in extra_idx_l: 
            ax.get_yticklabels()[idx].set_color("darkred")            

        ax.set_title(title, y=1.02)
        ax.set_xlabel('')
        ax.set_ylabel(ylabel, labelpad=20)

        # Manage frequency of xticks & make sure the last one always visible
        if xticks_freq:
            start, end = ax.get_xlim()
            t_seq = np.append(np.arange(start, end-5, xticks_freq), end)
            ax.xaxis.set_ticks(t_seq)

        #ax.tick_params(direction='out', pad=20)
        ax.tick_params(axis='x', rotation=45, pad=20)
        plt.ylim(0,100)

        return(fig)

    # Plot the probability integral transform test
    def pit_plot(self,
                 xlabel='Quantiles',
                 ylabel='Cumulative probability',
                 title=('Out-of-sample conditional density:'
                        ' Probability Integral Transform (PIT) test')):
        
        # Data work (note that the pit are computed by default)
        support = np.arange(0,1, 0.01)
        pits = self.dfor['pit'].dropna().copy()

        # Compute the ecdf on the pits
        ecdf = ECDF(pits)

        # Fit it on the support
        pit_line = ecdf(support)

        # Compute the KS statistics (in case of need)
        ks_stats = stats.kstest(self.dfor['pit'].dropna(), 'uniform')

        # Confidence intervals based on Rossi and Shekopysan JoE 2019
        ci_u = [x+1.61*len(pits)**(-0.5) for x in support]
        ci_l = [x-1.61*len(pits)**(-0.5) for x in support]

        #new_title = (f'{title} \n Kolmogorov-Smirnov Test of Uniformity')
        
        # Prepare the plots
        fig, ax = plt.subplots(1)
                
        ax.plot(support, pit_line, color='blue',
                label='Out-of-sample empirical CDF',
                lw=2)
        ax.plot(support, support, color='red', label='Theoretical CDF')
        ax.plot(support, ci_u, color='red', label='1 percent critical values',
                linestyle='dashed')
        ax.plot(support, ci_l, color='red', linestyle='dashed')
        ax.legend()
        ax.set_xlabel(xlabel, labelpad=20)
        ax.set_ylabel(ylabel, labelpad=20)
        ax.set_title(title, y=1.02)

        return(fig)


    # Fan chart
    def plot_fan_chart(self,
                       title=None,
                       ylabel='',
                       xticks_freq=None):
        """ 
        Plot a fan chart of the conditional density
        Colors focus on the extreme points instead than the median

        xticks_freq: None or int, default None
            Frequency of xticks: e.g., 2 is 1 out of 2. None for plt default

        """

        # Select the quantiles
        qfc_l = [0.05, 0.1, 0.25, 0.5, 0.75, 0.9, 0.95]
        qfc_labels_l = [f'cond_quant_{x}' for x in qfc_l]

        dfc = self.dfor[['true_val'] + qfc_labels_l].copy()

        dfc.columns = [x.replace('cond_quant_', '') for x in dfc.columns]


        # Prepare the plot
        fig, ax = plt.subplots(1, 1)
        
        # Plot each quantile values
        ax.plot(dfc['0.05'], linestyle=':', color='black', label='5th')
        ax.plot(dfc['0.1'], linestyle='-.', color='black', label='10th')
        ax.plot(dfc['0.25'], linestyle='--', color='black', label='25th')
        ax.plot(dfc['0.5'], linestyle='-', color='black', label='Median',
                lw=2)
        ax.plot(dfc['0.75'], linestyle='--', color='black', label='75th')
        ax.plot(dfc['0.9'], linestyle='-.', color='black', label='90th')
        ax.plot(dfc['0.95'], linestyle=':', color='black', label='95th')

        # Fill the colors between the lines with different transparency level
        ax.fill_between(dfc.index, dfc['0.05'], dfc['0.1'],
                        color='darkred', alpha=0.75)
        ax.fill_between(dfc.index, dfc['0.9'], dfc['0.95'],
                        color='darkred', alpha=0.75)
        ax.fill_between(dfc.index, dfc['0.1'], dfc['0.25'],
                        color='red', alpha=0.4)
        ax.fill_between(dfc.index, dfc['0.75'], dfc['0.9'],
                        color='red', alpha=0.4)
        ax.fill_between(dfc.index, dfc['0.25'], dfc['0.5'],
                        color='blue', alpha=0.15)
        ax.fill_between(dfc.index, dfc['0.5'], dfc['0.75'],
                        color='blue', alpha=0.15)  

        if title==None:
            t1 = f'Fan chart of predictive {self.depvar}'
            ttl = t1 + '\n 1, 5, 10, 25, 50, 75, 90, 95 Conditional Quantiles'
        else:
            ttl=title
        ax.set_title(ttl, y=1.02)
        ax.set_xlabel('')
        ax.set_ylabel(ylabel, labelpad=20)

        # Manage frequency of xticks & make sure the last one always visible
        if xticks_freq:
            start, end = ax.get_xlim()
            t_seq = np.append(np.arange(start, end-5, xticks_freq), end)            
            ax.xaxis.set_ticks(t_seq)

        ax.tick_params(axis='x', rotation=45)

                
        return(fig)

